@@ -8,15 +8,17 @@ import time
 import uuid
 import requests
 import traceback
-import sys
 
 from ..config import NERD_COMPUTE_ENDPOINT, DEBUG_MODE
 from ..spinner import Spinner
 from ..utils import debug_print, check_job_manually
+from .storage.data_size_utils import is_large_data
+from .storage.data_parsing_utils import parse_fetched_data
 from .import_utils import extract_imports, _extract_used_names, _filter_imports_by_usage, _get_stdlib_modules
 from .job import _active_jobs
 from .auth import get_api_key
-from .storage import is_large_data, upload_nerd_cloud_storage
+from .storage.storage import upload_nerd_cloud_storage
+from .on_server.on_server import generate_cloud_code, deserialize_arg
 
 def cloud_compute(cores=8, timeout=1800):
     """
@@ -30,6 +32,30 @@ def cloud_compute(cores=8, timeout=1800):
         A decorated function that will run in the cloud instead of locally
     """
     def decorator(func):
+        # Create a wrapper function that automatically handles data references
+        @functools.wraps(func)
+        def auto_reference_wrapper(*args, **kwargs):
+            # Import locally to avoid circular imports
+            from .helpers import fetch_nerd_data_reference
+
+            # Recursively check and resolve any references in args/kwargs
+            def resolve_references(obj):
+                if isinstance(obj, dict) and "__nerd_data_reference" in obj:
+                    return fetch_nerd_data_reference(obj)
+                elif isinstance(obj, list):
+                    return [resolve_references(item) for item in obj]
+                elif isinstance(obj, dict):
+                    return {k: resolve_references(v) for k, v in obj.items()}
+                return obj
+
+            # Resolve any references in the arguments
+            resolved_args = [resolve_references(arg) for arg in args]
+            resolved_kwargs = {k: resolve_references(v) for k, v in kwargs.items()}
+
+            # Call the original function with resolved arguments
+            return func(*resolved_args, **resolved_kwargs)
+
+        # Create a cloud wrapper that handles upload/download
         @functools.wraps(func)
         def wrapper(*args, **kwargs):
             # Check if API_KEY is set before proceeding
@@ -43,9 +69,53 @@ def cloud_compute(cores=8, timeout=1800):
 
             print(f"🚀 Running {func.__name__} on cloud server with {cores} cores...")
 
-            # Handle large binary files by using the default endpoint that doesn't require special permissions
-            # Skip the _send_compute_job approach since it's causing 403 errors
-            debug_print("Using standard cloud compute approach for all data")
+            # Check for and handle large binary data arguments
+            modified_args = list(args)
+            for i, arg in enumerate(args):
+                # Use the extracted utility to check if data is large
+                if is_large_data(arg, threshold_mb=3):
+                    debug_print(f"Large argument detected at index {i}")
+                    # Upload to cloud storage
+                    upload_result = upload_nerd_cloud_storage(arg)
+                    # Replace with reference object that can be recognized in the function
+                    modified_args[i] = {
+                        "__nerd_data_reference": upload_result["dataId"],
+                        "__nerd_s3_uri": upload_result.get("s3Uri", ""),
+                        "__nerd_size_mb": upload_result.get("sizeMB", "")
+                    }
+                    debug_print(f"Uploaded large arg to cloud storage: {upload_result['dataId']}")
+
+            # Check for large binary data in keyword arguments
+            modified_kwargs = {}
+            for key, value in kwargs.items():
+                # Use the extracted utility to check if data is large
+                if is_large_data(value, threshold_mb=3):
+                    debug_print(f"Large kwarg detected: {key}")
+                    # Upload to cloud storage
+                    upload_result = upload_nerd_cloud_storage(value)
+                    # Replace with reference object
+                    modified_kwargs[key] = {
+                        "__nerd_data_reference": upload_result["dataId"],
+                        "__nerd_s3_uri": upload_result.get("s3Uri", ""),
+                        "__nerd_size_mb": upload_result.get("sizeMB", "")
+                    }
+                    debug_print(f"Uploaded large kwarg to cloud storage: {upload_result['dataId']}")
+                else:
+                    modified_kwargs[key] = value
+
+            # After processing all kwargs, update args/kwargs if any upload occurred
+            uploaded_arg = any(
+                isinstance(a, dict) and "__nerd_data_reference" in a
+                for a in modified_args
+            )
+            uploaded_kwarg = any(
+                isinstance(v, dict) and "__nerd_data_reference" in v
+                for v in modified_kwargs.values()
+            )
+            if uploaded_arg or uploaded_kwarg:
+                debug_print("Using modified arguments with cloud storage references")
+                args = tuple(modified_args)
+                kwargs = modified_kwargs
 
             # Step 1: Get the actual code of your function
             source = inspect.getsource(func)
@@ -117,32 +187,11 @@ def cloud_compute(cores=8, timeout=1800):
             import_block = "\n".join(all_imports)
             debug_print(f"Generated import block:\n{import_block}")
 
-            # Add a fallback universal import for edge cases
-            universal_imports = """
-# Fallback imports for common libraries
-try:
-    import numpy as np
-except ImportError:
-    pass
-try:
-    import pandas as pd
-except ImportError:
-    pass
-try:
-    import matplotlib.pyplot as plt
-except ImportError:
-    pass
-"""
-            import_block = universal_imports + import_block
-
-            # Step 2: Package up all the data your function needs
-            spinner = Spinner("Packaging function and data for cloud execution...")
-            spinner.start()
-
+            # Serialize the arguments for cloud execution
             def serialize_with_chunking(obj):
                 try:
-                    # Special handling for large binary data
-                    if isinstance(obj, bytes) and len(obj) > 10 * 1024 * 1024:  # > 10MB
+                    # Special handling for large binary data that should go to S3 directly
+                    if isinstance(obj, bytes) and len(obj) > 5 * 1024 * 1024:  # > 5MB
                         size_mb = len(obj) / (1024 * 1024)
                         debug_print(f"Large binary data detected: {size_mb:.2f} MB")
                         # Upload to cloud storage
@@ -158,9 +207,12 @@ except ImportError:
                             }
                         }
 
-                    # Regular serialization for normal-sized data
+                    # Standard serialization for everything else:
+                    # 1. Pickle the object to binary form
                     pickled = pickle.dumps(obj)
+                    # 2. Compress the pickled data
                     compressed = zlib.compress(pickled)
+                    # 3. Base64 encode the compressed data
                     encoded = base64.b64encode(compressed).decode('utf-8')
                     return {'type': 'data', 'value': encoded}
                 except Exception as e:
@@ -177,99 +229,13 @@ except ImportError:
             for key, value in kwargs.items():
                 serialized_kwargs[key] = serialize_with_chunking(value)
 
-            cloud_code = """
-import pickle
-import base64
-import zlib
-import json
-import time
-import os
-import traceback
-import sys
-
-# This function unpacks the data we sent
-def deserialize_arg(arg_data):
-    if arg_data['type'] == 'data':
-        try:
-            return pickle.loads(zlib.decompress(base64.b64decode(arg_data['value'])))
-        except Exception as e:
-            print(f"Error deserializing: {e}")
-            return arg_data['value']
-    else:
-        return arg_data['value']
-
-# Debug function to get environment variables
-def debug_env():
-    env_vars = {}
-    for key in os.environ:
-        env_vars[key] = os.environ.get(key, 'NOT_SET')
-    return env_vars
-
-print(f"Cloud environment: {json.dumps(debug_env())}")
-
-# Auto-imported modules extracted from function code
-""" + import_block + """
-
-# Your original function is copied below (without the decorator)
-""" + source + """
-
-# Unpack all the arguments
-args = []
-for arg_data in """ + str(serialized_args) + """:
-    args.append(deserialize_arg(arg_data))
-
-# Unpack all the keyword arguments
-kwargs = {}
-for key, arg_data in """ + str(serialized_kwargs) + """.items():
-    kwargs[key] = deserialize_arg(arg_data)
-
-try:
-    print(f"Starting cloud execution of """ + func.__name__ + """...")
-    result = """ + func.__name__ + """(*args, **kwargs)
-    print(f"Function execution completed successfully")
-
-    try:
-        print("Packaging results to send back...")
-        result_pickled = pickle.dumps(result)
-        result_compressed = zlib.compress(result_pickled)
-        result_encoded = base64.b64encode(result_compressed).decode('utf-8')
-        print(f"Results packaged (size: {len(result_encoded)} characters)")
-
-        print("RESULT_MARKER_BEGIN")
-        print(f'{{"result_size": {len(result_encoded)}, "result": "{result_encoded}"}}')
-        print("RESULT_MARKER_END")
-
-        with open('/tmp/result.json', 'w') as f:
-            f.write(f'{{"result_size": {len(result_encoded)}, "result": "{result_encoded}"}}')
-        print("Saved result to /tmp/result.json")
-
-        try:
-            alternative_paths = ['/mnt/data/result.json', './result.json']
-            for alt_path in alternative_paths:
-                try:
-                    with open(alt_path, 'w') as f:
-                        f.write(f'{{"result_size": {len(result_encoded)}, "result": "{result_encoded}"}}')
-                    print(f"Also saved result to {alt_path}")
-                except:
-                    pass
-        except Exception as e:
-            print(f"Error saving to alternative paths: {e}")
-
-        sys.stdout.flush()
-        time.sleep(5)
-    except Exception as e:
-        print(f"Error packaging results: {e}")
-        print(traceback.format_exc())
-        raise
-except Exception as e:
-    print(f"EXECUTION ERROR: {e}")
-    print(traceback.format_exc())
-    raise
-"""
+            # Get cloud code for execution from the on_server module
+            cloud_code = generate_cloud_code(source, func.__name__, serialized_args, serialized_kwargs, import_block)
             job_id = str(uuid.uuid4())
             debug_print(f"Job ID: {job_id}")
 
-            spinner.update_message(f"Sending {func.__name__} to cloud server...")
+            spinner = Spinner("Packaging function and data for cloud execution...")
+            spinner.start()
 
             headers = {
                 "Content-Type": "application/json",
@@ -460,6 +426,7 @@ except Exception as e:
                                             if "result" in result_json and "result_size" in result_json:
                                                 try:
                                                     encoded_result = result_json["result"]
+                                                    # Using standard pickle/zlib/base64 deserialization as that's how the result is encoded
                                                     decoded_result = pickle.loads(zlib.decompress(base64.b64decode(encoded_result)))
                                                     spinner.update_message(f"Cloud computation completed in {int(elapsed)}s")
                                                     spinner.stop()
@@ -620,18 +587,9 @@ def _send_compute_job(function_info, args, kwargs, timeout=3600):
     # Process positional arguments
     serialized_args = []
     for i, arg in enumerate(args):
-        # Force large file handling for binary data > 10MB regardless of is_large_data result
-        if isinstance(arg, bytes) and len(arg) > 10 * 1024 * 1024:
-            debug_print(f"Forcing large file handling for binary arg {i}")
-            large_args.append(i)
-            upload_result = upload_nerd_cloud_storage(arg)
-            serialized_args.append({
-                "data_reference": {
-                    "type": "cloud_storage",
-                    "dataId": upload_result["dataId"]
-                }
-            })
-        elif is_large_data(arg):
+        # Use the extracted utility to check if data is large
+        if is_large_data(arg, threshold_mb=3):
+            debug_print(f"Large argument detected at index {i}")
             large_args.append(i)
             upload_result = upload_nerd_cloud_storage(arg)
             serialized_args.append({
@@ -646,18 +604,9 @@ def _send_compute_job(function_info, args, kwargs, timeout=3600):
     # Process keyword arguments
     serialized_kwargs = {}
     for key, value in kwargs.items():
-        # Force large file handling for binary data > 10MB
-        if isinstance(value, bytes) and len(value) > 10 * 1024 * 1024:
-            debug_print(f"Forcing large file handling for binary kwarg {key}")
-            large_kwargs[key] = value
-            upload_result = upload_nerd_cloud_storage(value)
-            serialized_kwargs[key] = {
-                "data_reference": {
-                    "type": "cloud_storage",
-                    "dataId": upload_result["dataId"]
-                }
-            }
-        elif is_large_data(value):
+        # Use the extracted utility to check if data is large
+        if is_large_data(value, threshold_mb=3):
+            debug_print(f"Large kwarg detected: {key}")
             large_kwargs[key] = value
             upload_result = upload_nerd_cloud_storage(value)
             serialized_kwargs[key] = {
